package com.loafle.overflow.container.configuration;

import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.Arrays;
import java.util.List;

import com.loafle.commons.rpc.protocol.RPCServerCodec;
import com.loafle.commons.rpc.protocol.json.JSONRPCServerCodec;
import com.loafle.commons.rpc.registry.RPCRegistry;
import com.loafle.commons.server.Server;
import com.loafle.commons.server.socket.handler.codec.SocketServerProtocolHandler;
import com.loafle.overflow.container.Container;
import com.loafle.overflow.container.server.handler.RPCServerHandler;

import org.codehaus.jackson.map.ObjectMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;

import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.ServerChannel;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.logging.LogLevel;
import io.netty.handler.logging.LoggingHandler;

/**
 * ContainerConfiguration
 */
@Configuration
@ComponentScan(basePackages = { "com.loafle.overflow" })
@PropertySource({ "classpath:netty.properties" })
public class ContainerConfiguration {
  @Autowired(required = false)
  @Qualifier(Container.PIPELINE_CHANNEL_HANDLERS)
  private List<ChannelHandler> pipelineChannelHandlers;

  @Autowired()
  private RPCServerHandler rpcServerHandler;

  @Autowired(required = false)
  private ObjectMapper objectMapper;

  @Bean(Server.CHANNEL_CLASS)
  public Class<? extends ServerChannel> channelClass() {
    return NioServerSocketChannel.class;
  }

  @Bean(Server.CHANNEL_OPTIONS)
  public List<Server.ChannelOptionItem<?>> channelOptions() {
    return Arrays.asList(new Server.ChannelOptionItem<>(ChannelOption.SO_BACKLOG, 100));
  }

  @Bean(Server.CHANNEL_HANDLER)
  public ChannelHandler channelHandler() {
    return new LoggingHandler(LogLevel.INFO);
  }

  @Bean(Container.RPC_SERVER_CODEC)
  public RPCServerCodec rpcServerCodec() {
    return new JSONRPCServerCodec(this.objectMapper);
  }

  @Bean({Container.RPC_INVOKER, Container.RPC_REGISTRY})
  public RPCRegistry rpcRegistry() {
    return new RPCRegistry();
  }

  @Bean(Server.CHANNEL_INITIALIZER)
  public ChannelInitializer<?> channelInitializer() {
    return new ChannelInitializer<SocketChannel>() {
      @Override
      public void initChannel(SocketChannel ch) throws Exception {
        ChannelPipeline cp = ch.pipeline();
        cp.addLast(new SocketServerProtocolHandler(true));
        cp.addLast(rpcServerHandler);
        if (null != pipelineChannelHandlers) {
          for (ChannelHandler channelHandler : pipelineChannelHandlers) {
            cp.addLast(channelHandler);
          }
        }
      }
    };
  }

  @Bean(Server.SOCKET_ADDRESS)
  public SocketAddress address() {
    return new InetSocketAddress("127.0.0.1", 60000);
  }
}