package com.loafle.overflow.container.server.handler;

import com.loafle.commons.rpc.RPCException;
import com.loafle.commons.rpc.protocol.RPCServerCodec;
import com.loafle.commons.rpc.protocol.RPCServerRequestCodec;
import com.loafle.commons.rpc.registry.RPCInvoker;
import com.loafle.commons.server.socket.handler.codec.SocketFrame;
import com.loafle.commons.server.socket.handler.codec.TextSocketFrame;
import com.loafle.overflow.container.Container;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Component;

import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;

/**
 * RPCDecoder
 */
@Component
public class RPCServerHandler extends SimpleChannelInboundHandler<SocketFrame> {
  private static final Logger logger = LoggerFactory.getLogger(RPCServerHandler.class);

  @Autowired
  @Qualifier(Container.RPC_SERVER_CODEC)
  private RPCServerCodec serverCodec;
  @Autowired()
  @Qualifier(Container.RPC_INVOKER)
  private RPCInvoker rpcInvoker;

  private ChannelHandlerContext ctx;

  public RPCServerHandler() {
  }

  @Override
  protected void channelRead0(ChannelHandlerContext ctx, SocketFrame frame) throws Exception {
    if (frame instanceof TextSocketFrame) {
      ByteBuf buff = null;
      byte[] requestBytes = null;
      RPCServerRequestCodec requestCodec = null;
      Object reply = null;
      RPCException error = null;
      try {
        buff = frame.content();
        requestBytes = new byte[buff.readableBytes()];
        buff.getBytes(buff.readerIndex(), requestBytes);

        requestCodec = this.serverCodec.request(requestBytes);
        reply = this.rpcInvoker.invoke(requestCodec);

      } catch (RPCException e) {
        logger.debug("Error", e);
        error = e;
      }
      if (!requestCodec.hasResponse()) {
        return;
      }
      byte[] responseBytes = requestCodec.response(reply, error);
      ctx.channel().writeAndFlush(responseBytes);
    } else {
      throw new UnsupportedOperationException(String.format("unsupported frame type: %s", frame.getClass().getName()));
    }
  }

  public void send(String method, Object... params) {
    if (null == this.ctx || !this.ctx.channel().isOpen()) {
      logger.warn("ChannelHandlerContext is not valid");
      return;
    }
    try {
      byte[] notificationBytes = this.serverCodec.notification(method, params);
      this.ctx.channel().writeAndFlush(notificationBytes);
    } catch (RPCException e) {
      logger.error("cannot create notification", e);
    }
  }

  @Override
  public void channelRegistered(ChannelHandlerContext ctx) throws Exception {
    // Initialize early if channel is active already.
    if (ctx.channel().isActive()) {
      this.ctx = ctx;
    }
    super.channelRegistered(ctx);
  }

  @Override
  public void channelActive(ChannelHandlerContext ctx) throws Exception {
    this.ctx = ctx;
    super.channelActive(ctx);
  }

  /**
   * Calls {@link ChannelHandlerContext#fireChannelInactive()} to forward
   * to the next {@link ChannelInboundHandler} in the {@link ChannelPipeline}.
   *
   * Sub-classes may override this method to change behavior.
   */
  @Override
  public void channelInactive(ChannelHandlerContext ctx) throws Exception {
    this.ctx = null;
    super.channelInactive(ctx);
  }
}